use std::error::Error;
use std::sync::Arc;

use clap::Parser;
use futures::future::join_all;
use reqwest::Certificate;

use base::RemoteConfig;
use tcp_relay_client::relay_client::relay_client;

/// TCP relay client
#[derive(Parser, Debug, Clone)]
#[clap(author, version, about, long_about = None)]
pub struct Args {
    /// Access token
    #[clap(short, long)]
    pub token: String,

    /// Relay server
    #[clap(short, long, default_value = "http://127.0.0.1:8000")]
    pub relay_url: String,

    /// Listen address
    #[clap(short, long, default_value = "127.0.0.1")]
    pub listen_address: String,

    /// Optional root certificate to use for server authentication
    #[clap(short = 'c', long)]
    pub root_certificate: Option<String>,
}

async fn get_server_config(config: &Args, root_cert: &Option<Vec<u8>>) -> Result<RemoteConfig, Box<dyn Error>> {
    let url = format!("{}/config", config.relay_url);
    log::info!("Retrieving configuration on {}", url);

    let mut client = reqwest::Client::builder();

    // Specify root certificate, if any was specified in the command line
    if let Some(cert) = root_cert {
        client = client.add_root_certificate(Certificate::from_pem(cert)?);
    }

    let client = client.build().expect("Failed to build reqwest client");

    let req = client.get(url)
        .header("Authorization", format!("Bearer {}", config.token))
        .send()
        .await?;
    if req.status().as_u16() != 200 {
        log::error!("Could not retrieve configuration! (got status {})", req.status());
        std::process::exit(2);
    }

    Ok(req.json::<RemoteConfig>().await?)
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    env_logger::init_from_env(env_logger::Env::new().default_filter_or("info"));

    let args: Args = Args::parse();
    let args = Arc::new(args);

    let root_cert = args.root_certificate.as_ref().map(|c| std::fs::read(c)
        .expect("Failed to read root certificate!"));

    // Get server relay configuration (fetch the list of port to forward)
    let conf = get_server_config(&args, &root_cert).await?;

    // Start to listen port
    let mut handles = vec![];
    for port in conf {
        let listen_address = format!("{}:{}", args.listen_address, port.port);

        let h = tokio::spawn(relay_client(
            format!("{}/ws?id={}&token={}",
                    args.relay_url, port.id, urlencoding::encode(&args.token))
                .replace("http", "ws"),
            listen_address,
            root_cert.clone(),
        ));
        handles.push(h);
    }

    join_all(handles).await;

    Ok(())
}